from fastmcp import FastMCP
from fastmcp.server.auth.providers.workos import AuthKitProvider
from fastmcp.prompts.prompt import Message
from fastmcp.resources import ResourceTemplate
from starlette.requests import Request
from starlette.responses import JSONResponse

from mqtt_mcp.mqtt_client import AsyncMQTTClient
from mqtt_mcp.settings import Settings


settings = Settings()


class MQTTMCP(FastMCP):
    def __init__(self, **kwargs):
        super().__init__(
            name="MQTT MCP Server",
            auth=(
                AuthKitProvider(
                    authkit_domain=settings.auth.domain, base_url=settings.auth.url
                )
                if settings.auth.domain and settings.auth.url
                else None
            ),
            **kwargs,
        )

        self.add_template(
            ResourceTemplate.from_function(
                fn=self.receive_message, uri_template="mqtt://{host}:{port}/{topic*}"
            )
        )

        self.tool(
            self.receive_message,
            annotations={
                "title": "Receive Message",
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        )

        self.tool(
            self.publish_message,
            annotations={
                "title": "Publish Message",
                "readOnlyHint": False,
                "openWorldHint": True,
            },
        )

        self.prompt(self.mqtt_error, name="mqtt_error", tags={"mqtt", "error"})
        self.prompt(self.mqtt_help, name="mqtt_help", tags={"mqtt", "help"})

        self.custom_route("/health", methods=["GET"])(self.health_check)

    async def receive_message(
        self,
        topic: str,
        host: str = settings.mqtt.host,
        port: int = settings.mqtt.port,
        username: str | None = settings.mqtt.username,
        password: str | None = settings.mqtt.password,
        timeout: int = 60,
    ) -> str:
        """Receives a message published to the specified topic, if any."""
        try:
            async with AsyncMQTTClient(host, port, username, password) as client:
                return await client.receive(topic, timeout)
        except Exception as e:
            raise RuntimeError(f"{e}") from e

    async def publish_message(
        self,
        topic: str,
        message: str,
        host: str = settings.mqtt.host,
        port: int = settings.mqtt.port,
        username: str | None = settings.mqtt.username,
        password: str | None = settings.mqtt.password,
    ) -> str:
        """Publishes a message to the specified topic."""
        try:
            async with AsyncMQTTClient(host, port, username, password) as client:
                await client.publish(topic, message)
            return f"Publish to {topic} on {host}:{port} has succedeed"
        except Exception as e:
            raise RuntimeError(f"{e}") from e

    def mqtt_help(self) -> list[Message]:
        """Provides examples of how to use the MQTT MCP server."""
        return [
            Message("Here are examples of how to publish and receives messages:"),
            Message('Publish {"foo":"bar"} to topic "devices/foo" on 127.0.0.1:1883.'),
            Message(
                'Receive a message from topic "devices/bar", waiting up to 30 seconds.'
            ),
        ]

    def mqtt_error(self, error: str | None = None) -> list[Message]:
        """Asks the user how to handle an error."""
        return (
            [
                Message(f"ERROR: {error!r}"),
                Message("Would you like to retry, change parameters, or abort?"),
            ]
            if error
            else []
        )

    async def health_check(self, request: Request) -> JSONResponse:
        return JSONResponse({"status": "ok"})



import asyncio
import pytest
import pytest_asyncio
import threading

from fastmcp import Client
from pydantic import BaseModel

from mqtt_mcp.server import MQTTMCP


class Config(BaseModel):
    host: str = "127.0.0.1"
    port: int = 1883


async def _server_main(config: Config) -> None:
    # NOTE: nothing to do here but this may change in the future.
    await asyncio.Future()


@pytest.fixture(scope="session")
def server():
    config = Config()
    thread = threading.Thread(
        target=lambda: asyncio.run(_server_main(config)), daemon=True
    )
    thread.start()
    yield config


@pytest.fixture(scope="session")
def mcp():
    return MQTTMCP()


@pytest_asyncio.fixture
async def client(mcp):
    async with Client(mcp) as c:
        yield c


@pytest.fixture()
def cli(monkeypatch):
    async def dummy_run_async(self, transport):
        return

    monkeypatch.setattr(
        "mqtt_mcp.cli.MQTTMCP.run_async",
        dummy_run_async,
    )



import asyncio
import pytest

from starlette.requests import Request


@pytest.mark.asyncio
async def test_receive_message(server, mcp, client):
    """Test receive_message."""
    topic = "foo"
    message = '{"bar":123}'

    async def pub():
        await asyncio.sleep(1.0)
        return await client.call_tool(
            "publish_message",
            {
                "topic": topic,
                "message": message,
                "host": server.host,
                "port": server.port,
            },
        )

    sub = None
    async with asyncio.TaskGroup() as tg:
        sub = tg.create_task(
            client.call_tool(
                "receive_message",
                {
                    "topic": topic,
                    "host": server.host,
                    "port": server.port,
                    "timeout": 3,
                },
            )
        )
        tg.create_task(pub())

    result = sub.result()
    assert len(result.content) == 1
    assert result.content[0].text == message


@pytest.mark.asyncio
async def test_publish_message(server, mcp, client):
    """Test publish_message."""
    result = await client.call_tool(
        "publish_message",
        {
            "topic": "foo",
            "message": '{"bar":456}',
            "host": server.host,
            "port": server.port,
        },
    )
    assert len(result.content) == 1
    assert "succedeed" in result.content[0].text


@pytest.mark.asyncio
async def test_help_prompt(mcp, client):
    """Test help prompt."""
    result = await client.get_prompt("mqtt_help", {})
    assert len(result.messages) == 3


@pytest.mark.asyncio
async def test_error_prompt(mcp, client):
    """Test error prompt."""
    result = await client.get_prompt("mqtt_error", {"error": "Could not read data"})
    assert len(result.messages) == 2


@pytest.mark.asyncio
async def test_health_check(mcp):
    response = await mcp.health_check(
        Request(
            {
                "type": "http",
                "method": "GET",
                "path": "/health",
                "headers": [],
            }
        )
    )
    assert response.status_code == 200



"""Async MQTT Client."""

import asyncio

import paho.mqtt.client as mqtt

from typing import Optional


class AsyncioHelper:
    """Integrate paho-mqtt socket callbacks with asyncio event loop."""

    def __init__(self, client):
        self.client = client

    def start_loop(self):
        self.client.loop_start()

    def stop_loop(self):
        self.client.loop_stop()


class AsyncMQTTClient:
    """Async MQTT client wrapper."""

    def __init__(
        self,
        host: str,
        port: int = 1883,
        username: str | None = None,
        password: str | None = None,
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, clean_session=True)
        if self.username and self.password:
            self.client.username_pw_set(self.username, self.password)
        self.future: Optional[asyncio.Future[str]] = None

    async def __aenter__(self) -> "AsyncMQTTClient":
        loop = asyncio.get_running_loop()
        self.helper = AsyncioHelper(self.client)

        # Set up connection callback
        connection_future = loop.create_future()

        def on_connect(client, userdata, flags, reason_code, properties):
            if reason_code == 0:
                loop.call_soon_threadsafe(connection_future.set_result, True)
            else:
                loop.call_soon_threadsafe(
                    connection_future.set_exception,
                    RuntimeError(f"MQTT connection failed with code {reason_code}"),
                )

        def on_disconnect(client, userdata, reason_code, properties=None, *args):
            # Handle both v1 and v2 callback signatures (v2 passes properties as 4th arg)
            # Additional args are ignored for compatibility
            if reason_code != 0:
                # Unexpected disconnection
                if not connection_future.done():
                    loop.call_soon_threadsafe(
                        connection_future.set_exception,
                        RuntimeError(f"MQTT disconnected with code {reason_code}"),
                    )

        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect

        # Start the client loop (Windows-compatible)
        self.helper.start_loop()

        # Connect to the broker
        self.client.connect(self.host, self.port, keepalive=60)

        # Wait for connection to be established (timeout after 5 seconds)
        try:
            await asyncio.wait_for(connection_future, timeout=5.0)
        except asyncio.TimeoutError:
            self.helper.stop_loop()
            raise RuntimeError(
                f"Failed to connect to MQTT broker at {self.host}:{self.port} (timeout)"
            )

        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.helper.stop_loop()
        self.client.disconnect()
        await asyncio.sleep(0.1)  # Give time for disconnect to complete

    async def receive(self, topic: str, timeout: int = 60, qos: int = 1) -> str:
        loop = asyncio.get_running_loop()
        self.future = loop.create_future()

        # Store original message callback if it exists
        original_on_message = getattr(self.client, "on_message", None)

        # Set up message callback that filters by topic
        def on_message(client, userdata, message):
            # Check if this message matches our topic
            if message.topic == topic and self.future and not self.future.done():
                try:
                    message_str = message.payload.decode()
                    loop.call_soon_threadsafe(self.future.set_result, message_str)
                except Exception as e:
                    if not self.future.done():
                        loop.call_soon_threadsafe(self.future.set_exception, e)
            # Call original callback if it exists (for chaining)
            elif original_on_message:
                original_on_message(client, userdata, message)

        # Set up subscription callback to know when subscription is confirmed
        subscription_future = loop.create_future()
        original_on_subscribe = getattr(self.client, "on_subscribe", None)

        def on_subscribe(client, userdata, mid, granted_qos, properties=None, *args):
            # Subscription confirmed
            if not subscription_future.done():
                loop.call_soon_threadsafe(subscription_future.set_result, True)
            # Call original callback if it exists
            if original_on_subscribe:
                original_on_subscribe(client, userdata, mid, granted_qos, properties)

        # Register callbacks
        self.client.on_subscribe = on_subscribe
        self.client.on_message = on_message

        # Subscribe to the topic
        result, mid = self.client.subscribe(topic, qos=qos)
        if result != mqtt.MQTT_ERR_SUCCESS:
            # Restore original callbacks on error
            if original_on_message:
                self.client.on_message = original_on_message
            if original_on_subscribe:
                self.client.on_subscribe = original_on_subscribe
            raise RuntimeError(f"Subscribe failed with code {result}")

        # Wait for subscription to be acknowledged (with timeout)
        try:
            await asyncio.wait_for(subscription_future, timeout=2.0)
        except asyncio.TimeoutError:
            # Subscription might still work even if ack is slow
            pass

        # Give a small delay to ensure subscription is active
        await asyncio.sleep(0.2)

        try:
            return await asyncio.wait_for(self.future, timeout)
        finally:
            # Restore original callbacks
            if original_on_message:
                self.client.on_message = original_on_message
            elif hasattr(self.client, "on_message"):
                # If no original, just remove our callback
                self.client.on_message = None
            if original_on_subscribe:
                self.client.on_subscribe = original_on_subscribe
            self.future = None

    async def publish(self, topic: str, message: str, qos: int = 1) -> None:
        result = self.client.publish(topic, message, qos=qos)
        if result.rc != mqtt.MQTT_ERR_SUCCESS:
            raise RuntimeError(f"Publish failed with code {result.rc}")
        # Wait for the publish to complete (especially for qos > 0)
        if qos > 0:
            result.wait_for_publish(timeout=5.0)
        # Give the event loop a chance to process the publish
        await asyncio.sleep(0.1)



from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict


class Auth(BaseModel):
    domain: str | None = None
    url: str | None = None


class MQTT(BaseModel):
    host: str = "127.0.0.1"
    port: int = 1883
    username: str | None = None
    password: str | None = None


class Settings(BaseSettings):
    auth: Auth = Auth()
    mqtt: MQTT = MQTT()
    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
        env_prefix="MQTT_MCP_",
    )
